#pragma once

#include <map>
#include <string>

#include "EntityBase.h"
#include "detail/AronTyped.h"
#include "detail/EntityContainerBase.h"
#include "detail/MaxHistorySize.h"


namespace armarx::armem::base
{

    /**
     * @brief Data of a provider segment containing multiple entities.
     */
    template <class _EntityT, class _Derived>
    class ProviderSegmentBase :
        public detail::EntityContainerBase<_EntityT, _EntityT, _Derived>,
        public detail::MaxHistorySize,
        public detail::AronTyped
    {
        using Base = detail::EntityContainerBase<_EntityT, _EntityT, _Derived>;

    public:

        using typename Base::DerivedT;
        using typename Base::ContainerT;

        using EntityT = _EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;
        using EntityInstanceT = typename EntitySnapshotT::EntityInstanceT;


    public:

        ProviderSegmentBase()
        {
        }

        ProviderSegmentBase(const std::string& name, aron::typenavigator::ObjectNavigatorPtr aronType = nullptr) :
            ProviderSegmentBase(MemoryID().withProviderSegmentName(name), aronType)
        {
        }
        ProviderSegmentBase(const std::string& name, const MemoryID parentID, aron::typenavigator::ObjectNavigatorPtr aronType = nullptr) :
            ProviderSegmentBase(parentID.withProviderSegmentName(name), aronType)
        {
        }
        ProviderSegmentBase(const MemoryID id, aron::typenavigator::ObjectNavigatorPtr aronType = nullptr) :
            Base(id),
            AronTyped(aronType)
        {
        }

        ProviderSegmentBase(const ProviderSegmentBase& other) = default;
        ProviderSegmentBase(ProviderSegmentBase&& other) = default;
        ProviderSegmentBase& operator=(const ProviderSegmentBase& other) = default;
        ProviderSegmentBase& operator=(ProviderSegmentBase&& other) = default;


        inline const std::string& name() const
        {
            return this->id().providerSegmentName;
        }
        inline std::string& name()
        {
            return const_cast<std::string&>(const_cast<const ProviderSegmentBase*>(this)->name());
        }


        inline const std::map<std::string, EntityT>& entities() const
        {
            return this->_container;
        }
        inline std::map<std::string, EntityT>& entities()
        {
            return const_cast<std::map<std::string, EntityT>&>(const_cast<const ProviderSegmentBase*>(this)->entities());
        }


        bool hasEntity(const std::string& name) const
        {
            return this->_container.count(name) > 0;
        }

        using Base::getEntity;
        const EntityT& getEntity(const MemoryID& id) const override
        {
            this->_checkContainerName(id.providerSegmentName, this->getKeyString());
            return getEntity(id.entityName);
        }

        EntityT& getEntity(const std::string& name)
        {
            return const_cast<EntityT&>(const_cast<const ProviderSegmentBase*>(this)->getEntity(name));
        }

        const EntityT& getEntity(const std::string& name) const
        {
            auto it = this->_container.find(name);
            if (it != this->_container.end())
            {
                return it->second;
            }
            else
            {
                throw error::MissingEntry("entity", name, getLevelName(), this->name());
            }
        }

        const EntityT* findEntity(const MemoryID& id) const override
        {
            this->_checkContainerName(id.providerSegmentName, this->getKeyString());
            auto it = this->_container.find(id.entityName);
            if (it != this->_container.end())
            {
                return &it->second;
            }
            else
            {
                return nullptr;
            }
        }

        /**
         * @brief Updates an entity's history.
         *
         * Missing entity entries are added before updating.
         */
        virtual MemoryID update(const EntityUpdate& update) override
        {
            this->_checkContainerName(update.entityID.providerSegmentName, this->name());

            EntityT* entity;
            auto it = this->_container.find(update.entityID.providerSegmentName);
            if (it == this->_container.end())
            {
                // Add entity entry.
                entity = &addEntity(update.entityID.entityName);
                entity->setMaxHistorySize(_maxHistorySize);
            }
            else
            {
                entity = &it->second;
            }
            // Update entity.
            return entity->update(update);
        }

        /// Add an empty entity with the given name.
        EntityT& addEntity(const std::string& name)
        {
            return addEntity(EntityT(name));
        }
        /// Copy and insert an entity.
        EntityT& addEntity(const EntityT& entity)
        {
            return addEntity(EntityT(entity));
        }
        /// Move and insert an entity.
        EntityT& addEntity(EntityT&& entity)
        {
            auto it = this->_container.emplace(entity.name(), std::move(entity)).first;
            it->second.id().setProviderSegmentID(this->id());
            return it->second;
        }


        /**
         * @brief Sets the maximum history size of container in this segment.
         * This affects all current container as well as new ones.
         * @see Entity::setMaxHistorySize()
         */
        void setMaxHistorySize(long maxSize) override
        {
            MaxHistorySize::setMaxHistorySize(maxSize);
            for (auto& [name, entity] : this->_container)
            {
                entity.setMaxHistorySize(maxSize);
            }
        }


        virtual bool equalsDeep(const ProviderSegmentBase& other) const
        {
            //std::cout << "ProviderSegment::equalsDeep" << std::endl;
            if (this->size() != other.size())
            {
                return false;
            }
            for (const auto& [key, value] : this->_container)
            {
                if (not other.hasEntity(key))
                {
                    return false;
                }

                if (not value.equalsDeep(other.getEntity(key)))
                {
                    return false;
                }
            }
            return true;
        }

        std::string getLevelName() const override
        {
            return "provider segment";
        }

        std::string getKeyString() const override
        {
            return this->name();
        }


    protected:

        virtual void _copySelf(DerivedT& other) const override
        {
            Base::_copySelf(other);
            other.aronType() = _aronType;
        }
        virtual void _copySelfEmpty(DerivedT& other) const override
        {
            Base::_copySelfEmpty(other);
            other.aronType() = _aronType;
        }

    };

}
