#include "ComponentPlugin.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error.h"

#include "MemoryToIceAdapter.h"

//#include <RobotAPI/libraries/armem/core/io/diskWriter/NlohmannJSON/NlohmannJSONDiskWriter.h>

namespace armarx::armem::server::plugins
{
    ComponentPlugin::~ComponentPlugin()
    {}


    void ComponentPlugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        ClientPlugin::postCreatePropertyDefinitions(properties);
        properties->topic(memoryListener, this->parent<ComponentPluginUser>().memoryListenerDefaultName);

        properties->optional(longTermMemoryDatabaseHost, this->parent<ComponentPluginUser>().longTermMemoryDatabaseHostDefault);
        properties->optional(longTermMemoryDatabaseUser, this->parent<ComponentPluginUser>().longTermMemoryDatabaseUserDefault);
        properties->optional(longTermMemoryDatabasePassword, this->parent<ComponentPluginUser>().longTermMemoryDatabasePasswordDefault);
    }


    void ComponentPlugin::postOnConnectComponent()
    {
        ComponentPluginUser& parent = this->parent<ComponentPluginUser>();

        if (isMemoryNameSystemEnabled() && parent.memoryNameSystem)
        {
            registerMemory(parent);
        }
        parent.iceMemory.setMemoryListener(memoryListener);

        // establishing connection to ltm and mongodb
        ltm::MongoDBConnectionManager::MongoDBSettings settings;
        settings.host = longTermMemoryDatabaseHost;
        settings.user = longTermMemoryDatabaseUser;
        settings.password = longTermMemoryDatabasePassword;

        parent.longtermMemory.reload(settings);
    }


    void ComponentPlugin::preOnDisconnectComponent()
    {
        ComponentPluginUser& parent = this->parent<ComponentPluginUser>();
        if (isMemoryNameSystemEnabled() && parent.memoryNameSystem)
        {
            removeMemory(parent);
        }
    }


    data::RegisterMemoryResult ComponentPlugin::registerMemory(ComponentPluginUser& parent)
    {
        data::RegisterMemoryInput input;
        input.name = parent.workingMemory.name();
        input.proxy = MemoryInterfacePrx::checkedCast(parent.getProxy());
        ARMARX_CHECK_NOT_NULL(input.proxy);
        data::RegisterMemoryResult result = parent.memoryNameSystem->registerMemory(input);
        if (result.success)
        {
            ARMARX_DEBUG << "Registered memory '" << input.name << "' in the Memory Name System (MNS).";
        }
        else
        {
            ARMARX_WARNING << "Failed to register this memory in the Memory Name System (MNS):"
                           << "\n" << result.errorMessage;
        }
        return result;
    }


    data::RemoveMemoryResult ComponentPlugin::removeMemory(ComponentPluginUser& parent)
    {
        data::RemoveMemoryResult result;
        try
        {
            data::RemoveMemoryInput input;
            input.name = parent.workingMemory.name();
            result = parent.memoryNameSystem->removeMemory(input);
            if (result.success)
            {
                ARMARX_DEBUG << "Removed memory '" << input.name << "' from the Memory Name System (MNS).";
            }
            else
            {
                ARMARX_WARNING << "Failed to remove this memory in the Memory Name System (MNS):"
                               << "\n" << result.errorMessage;
            }
        }
        catch (const Ice::NotRegisteredException&)
        {
            // It's ok, the MNS is gone.
            result.success = false;
            result.errorMessage = "Memory Name System is gone.";
        }
        return result;
    }

}


namespace armarx::armem::server
{
    ComponentPluginUser::ComponentPluginUser()
    {
        addPlugin(plugin);
    }


    // WRITING
    data::AddSegmentsResult ComponentPluginUser::addSegments(const data::AddSegmentsInput& input, const Ice::Current&)
    {
        bool addCoreSegmentOnUsage = false;
        return addSegments(input, addCoreSegmentOnUsage);
    }

    data::AddSegmentsResult ComponentPluginUser::addSegments(const data::AddSegmentsInput& input, bool addCoreSegments)
    {
        std::scoped_lock lock(workingMemoryMutex);
        data::AddSegmentsResult result = iceMemory.addSegments(input, addCoreSegments);
        return result;
    }


    data::CommitResult ComponentPluginUser::commit(const data::Commit& commitIce, const Ice::Current&)
    {
        std::scoped_lock lock(workingMemoryMutex);
        return iceMemory.commit(commitIce);
    }


    // READING
    armem::query::data::Result ComponentPluginUser::query(const armem::query::data::Input& input, const Ice::Current&)
    {
        std::scoped_lock lock(workingMemoryMutex);
        return iceMemory.query(input);
    }

    // LTM LOADING
    data::StoreResult ComponentPluginUser::store(const data::StoreInput& input, const Ice::Current&)
    {
        std::scoped_lock lock(workingMemoryMutex);
        std::scoped_lock lock2(longtermMemoryMutex);
        return iceMemory.store(input);
    }


    // LTM STORING
    armem::query::data::Result ComponentPluginUser::load(const armem::query::data::Input& input, const Ice::Current&)
    {
        std::scoped_lock lock(longtermMemoryMutex);
        return iceMemory.load(input);
    }

}
