/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL

// ArmarX
#include <RobotAPI/interface/aron.h>
#include <ArmarXCore/core/exceptions/Exception.h>
#include <RobotAPI/libraries/aron/core/Path.h>
#include <RobotAPI/libraries/aron/core/Descriptor.h>

namespace armarx::aron::error
{
    class AronException :
        public armarx::LocalException
    {
    public:
        AronException() = delete;
        AronException(const std::string& caller, const std::string& method, const std::string& reason) :
            LocalException(caller + "::" + method + ": " + reason + ".")
        {

        }

        AronException(const std::string& caller, const std::string& method, const std::string& reason, const Path& path) :
            LocalException(caller + "::" + method + ": " + reason + ". The path was: " + path.toString())
        {

        }
    };

    class NotImplementedYetException :
        public AronException
    {
    public:
        NotImplementedYetException() = delete;
        NotImplementedYetException(const std::string& caller, const std::string& method) :
            AronException(caller, method, "This method is not yet implemented!")
        {

        }
    };

    class AronNotValidException :
        public AronException
    {
    public:
        AronNotValidException() = delete;
        AronNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const data::AronDataPtr& data) :
            AronException(caller, method, reason + ". The ice_id of the data was: " + data->ice_id())
        {

        }

        AronNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const data::AronDataPtr& data, const Path& path) :
            AronException(caller, method, reason + ". The ice_id of the data was: " + data->ice_id(), path)
        {

        }

        AronNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const type::AronTypePtr& type) :
            AronException(caller, method, reason + ". The ice_id of the type was: " + type->ice_id())
        {

        }

        AronNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const type::AronTypePtr& type, const Path& path) :
            AronException(caller, method, reason + ". The ice_id of the type was: " + type->ice_id(), path)
        {

        }
    };

    class DescriptorNotValidException :
        public AronException
    {
    public:
        DescriptorNotValidException() = delete;
        DescriptorNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const type::Descriptor& type) :
            AronException(caller, method, reason + ". The name of the type-enum was: " + type::DESCRIPTOR_TO_STRING(type))
        {

        }

        DescriptorNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const type::Descriptor& type, const Path& path) :
            AronException(caller, method, reason + ". The name of the type-enum was: " + type::DESCRIPTOR_TO_STRING(type), path)
        {

        }

        DescriptorNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const data::Descriptor& data) :
            AronException(caller, method, reason + ". The name of the data-enum was: " + data::DESCRIPTOR_TO_STRING(data))
        {

        }

        DescriptorNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const data::Descriptor& data, const Path& path) :
            AronException(caller, method, reason + ". The name of the data-enum was: " + data::DESCRIPTOR_TO_STRING(data), path)
        {

        }
    };

    class StringNotValidException :
        public AronException
    {
    public:
        StringNotValidException() = delete;
        StringNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const std::string& input) :
            AronException(caller, method, reason + ". Got: " + input)
        {

        }

        StringNotValidException(const std::string& caller, const std::string& method, const std::string& reason, const std::string& input, const Path& path) :
            AronException(caller, method, reason + ". Got: " + input, path)
        {

        }
    };

    class IndexNotValidException :
        public AronException
    {
    public:
        IndexNotValidException() = delete;
        IndexNotValidException(const std::string& caller, const std::string& method, const std::string& reason, int real, int expected) :
            AronException(caller, method, reason + ". The requested index is: " + std::to_string(real) + ". The maximum index was: " + std::to_string(expected))
        {

        }

        IndexNotValidException(const std::string& caller, const std::string& method, const std::string& reason, int real, int expected, const Path& path) :
            AronException(caller, method, reason + ". The requested index is: " + std::to_string(real) + ". The maximum index was: " + std::to_string(expected), path)
        {

        }
    };

    class SizeNotValidException :
        public AronException
    {
    public:
        SizeNotValidException() = delete;
        SizeNotValidException(const std::string& caller, const std::string& method, const std::string& reason, int real, int expected) :
            AronException(caller, method, reason + ". The measured size is: " + std::to_string(real) + ". The expected size was: " + std::to_string(expected))
        {

        }

        SizeNotValidException(const std::string& caller, const std::string& method, const std::string& reason, int real, int expected, const Path& path) :
            AronException(caller, method, reason + ". The measured size is: " + std::to_string(real) + ". The expected size was: " + std::to_string(expected), path)
        {

        }
    };
}
