/*
* This file is part of ArmarX.
*
* Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
* Karlsruhe Institute of Technology (KIT), all rights reserved.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL

// Header
#include <RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/serializer/Serializer.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/serializer/SerializerFactory.h>

namespace armarx::aron::cppcodegenerator
{
    // constantes
    const std::string Serializer::ARON_DATA_NAME = simox::meta::get_type_name(typeid(armarx::aron::data::AronData));
    const std::string Serializer::ARON_DATA_PTR_NAME = Serializer::ARON_DATA_NAME + "::PointerType";

    const std::string Serializer::ARON_TYPE_NAME = simox::meta::get_type_name(typeid(armarx::aron::type::AronType));
    const std::string Serializer::ARON_TYPE_PTR_NAME = Serializer::ARON_TYPE_NAME + "::PointerType";

    const std::map<std::string, std::string> Serializer::RESOLVE_TYPENAMES =
    {
        {"string", "std::string"},
    };

    const std::map<std::string, std::string> Serializer::ESCAPE_ACCESSORS =
    {
        {"->", "_ptr_"},
        {".", "_dot_"},
        {"[", "_lbr_"},
        {"]", "_rbr_"},
    };

    const SerializerFactoryPtr Serializer::FACTORY = SerializerFactoryPtr(new SerializerFactory());


    // static methods
    std::string Serializer::ResolveCppTypename(const std::string& enteredTypeName)
    {
        auto i = Serializer::RESOLVE_TYPENAMES.find(enteredTypeName);
        if (i != Serializer::RESOLVE_TYPENAMES.end())
        {
            return i->second;
        }
        return enteredTypeName;
    }

    std::vector<std::string> Serializer::ResolveCppTypenames(const std::vector<std::string>& enteredTypeNames)
    {
        std::vector<std::string> ret;
        for (const auto& t : enteredTypeNames)
        {
            ret.push_back(Serializer::ResolveCppTypename(t));
        }
        return ret;
    }

    std::string Serializer::EscapeAccessor(const std::string& accessor)
    {
        std::string escaped_accessor = accessor;
        for (const auto& [key, value] : ESCAPE_ACCESSORS)
        {
            escaped_accessor = simox::alg::replace_all(escaped_accessor, key, value);
        }
        return escaped_accessor;
    }

    std::string Serializer::UnescapeAccessor(const std::string& accessor)
    {
        std::string unescaped_accessor = accessor;
        for (const auto& [key, value] : ESCAPE_ACCESSORS)
        {
            unescaped_accessor = simox::alg::replace_all(unescaped_accessor, value, key);
        }
        return unescaped_accessor;
    }

    std::string Serializer::ExtractCppTypename(const typenavigator::NavigatorPtr& n)
    {
        SerializerPtr cpp = Serializer::FromAronTypeNaviagtorPtr(n);
        return cpp->getFullCppTypename();
    }

    std::vector<std::string> Serializer::ExtractCppTypenames(const std::vector<typenavigator::NavigatorPtr>& n)
    {
        std::vector<std::string> ret;
        for (const auto& typenav : n)
        {
            ret.push_back(Serializer::ExtractCppTypename(typenav));
        }
        return ret;
    }

    SerializerPtr Serializer::FromAronTypeNaviagtorPtr(const typenavigator::NavigatorPtr& n)
    {
        return FACTORY->create(n, n->getPath());
    }

    // constructors
    Serializer::Serializer(const std::string& cppName, const std::string& aronDataTypename, const std::string& aronTypeTypename, bool is_ptr) :
        is_ptr(is_ptr),
        cppTypename(cppName),
        aronDataTypename(aronDataTypename),
        aronTypeTypename(aronTypeTypename)
    {

    }

    std::string Serializer::getCoreCppTypename() const
    {
        return cppTypename;
    }

    std::string Serializer::getFullCppTypename() const
    {
        return is_ptr ? "std::shared_ptr<" + cppTypename + ">" : cppTypename;
    }

    std::string Serializer::getAronDataTypename() const
    {
        return aronDataTypename;
    }

    std::string Serializer::getAronDataPtrTypename() const
    {
        return getAronDataTypename() + "::PointerType";
    }

    std::string Serializer::getAronTypeTypename() const
    {
        return aronTypeTypename;
    }

    std::string Serializer::getAronTypePtrTypename() const
    {
        return getAronTypeTypename() + "::PointerType";
    }

    CppCtorPtr Serializer::toCtor(const std::string& name) const
    {
        CppCtorPtr c = CppCtorPtr(new CppCtor(name + "()"));
        std::vector<std::pair<std::string, std::string>> initList = this->getCtorInitializers("");
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("initialize();");
        b->appendBlock(this->getCtorBlock(""));
        c->addInitListEntries(initList);
        c->setBlock(b);

        return c;
    }

    CppMethodPtr Serializer::toResetMethod() const
    {
        std::stringstream doc;
        doc << "@brief reset() - This method resets all member variables to default. \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void reset() override", doc.str()));
        CppBlockPtr b = this->getResetBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toInitializeMethod() const
    {
        std::stringstream doc;
        doc << "@brief initialize() - This method initializes member variables according to the XML type description. \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void initialize() override", doc.str()));
        CppBlockPtr b = this->getInitializeBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toWriteInitialTypeMethod() const
    {
        std::stringstream doc;
        doc << "@brief writeType() - This method returns a new type from the class structure using a type writer implementation. This function is static. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("static void writeInitialType(armarx::aron::typeIO::WriterInterface& w)", doc.str()));
        CppBlockPtr b = this->getWriteInitialTypeBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toWriteCurrentTypeMethod() const
    {
        std::stringstream doc;
        doc << "@brief writeType() - This method returns a new type from the current class structure using a type writer implementation. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("void writeCurrentType(armarx::aron::typeIO::WriterInterface& w) const override", doc.str()));
        CppBlockPtr b = this->getWriteCurrentTypeBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toWriteMethod() const
    {
        std::stringstream doc;
        doc << "@brief write() - This method returns a new type from the member data types using a data writer implementation. \n";
        doc << "@param w - The writer implementation\n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void write(armarx::aron::dataIO::WriterInterface& w) const override", doc.str()));
        CppBlockPtr b = this->getWriteBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toReadMethod() const
    {
        std::stringstream doc;
        doc << "@brief read() - This method sets the struct members to new values given in a data reader implementation. \n";
        doc << "@param r - The reader implementation\n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("virtual void read(armarx::aron::dataIO::ReaderInterface& r) override", doc.str()));
        CppBlockPtr b = this->getReadBlock("");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr Serializer::toSpecializedDataWriterMethod(const std::string& returnname, const std::string& methodname, const std::string& writerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedDataWrite() - This method returns a new data from the member data types using a writer implementation. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod(returnname + " " + methodname + "() const", doc.str()));
        m->addLine(writerName + " writer;");
        m->addLine("this->write(writer);");
        m->addLine("return " + enforceConversion + "(writer.getResult());");
        return m;
    }

    CppMethodPtr Serializer::toSpecializedDataReaderMethod(const std::string& argumentname, const std::string& methodname, const std::string& readerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedDataRead() - This method sets the struct members to new values given in a reader implementation. \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("void " + methodname + "(const " + argumentname + "& input)", doc.str()));
        m->addLine(readerName + " reader(" + enforceConversion + "(input));");
        m->addLine("this->read(reader);");
        return m;
    }

    CppMethodPtr Serializer::toSpecializedInitialTypeWriterMethod(const std::string& returnname, const std::string& methodname, const std::string& writerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedTypeWrite() - This method returns a new type from the member data types using a writer implementation. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("static " + returnname + " " + methodname + "()", doc.str()));
        m->addLine(writerName + " writer;");
        m->addLine("writeInitialType(writer);");
        m->addLine("return " + enforceConversion + "(writer.getResult());");
        return m;
    }

    CppMethodPtr Serializer::toSpecializedCurrentTypeWriterMethod(const std::string& returnname, const std::string& methodname, const std::string& writerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedTypeWrite() - This method returns a new type from the current member data types using a writer implementation. \n";
        doc << "@return - the result of the writer implementation";

        CppMethodPtr m = CppMethodPtr(new CppMethod(returnname + " " + methodname + "()", doc.str()));
        m->addLine(writerName + " writer;");
        m->addLine("this->writeCurrentType(writer);");
        m->addLine("return " + enforceConversion + "(writer.getResult());");
        return m;
    }

    CppMethodPtr Serializer::toSpecializedCurrentTypeReaderMethod(const std::string& argumentname, const std::string& methodname, const std::string& readerName, const std::string& enforceConversion) const
    {
        std::stringstream doc;
        doc << "@brief specializedTypeRead() - This method sets the structure of the members to new values given in a reader implementation. \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("void " + methodname + "(const " + argumentname + "& input)", doc.str()));
        m->addLine(readerName + " reader(" + enforceConversion + "(input));");
        m->addLine("this->readType(reader);");
        return m;
    }

    CppMethodPtr Serializer::toEqualsMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator==() - This method checks whether all values equal another instance. \n";
        doc << "@param i - The other instance\n";
        doc << "@return - true, if all members are the same, false otherwise";

        CppMethodPtr m = CppMethodPtr(new CppMethod("bool operator==(const " + this->getFullCppTypename() + "& i) const", doc.str()));
        CppBlockPtr b = this->getEqualsBlock("", "i");
        b->addLine("return true;");
        m->setBlock(b);
        return m;
    }

}


