/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NavigatorWriter.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>



namespace armarx::aron::dataIO::writer
{

    Path NavigatorWriter::generatePath()
    {
        Path path;
        if (!wroteInitialStartContainer)
        {
            wroteInitialStartContainer = true;
        }
        else
        {
            auto parent_token = stack.top();
            path = Path(parent_token->getElement()->getPath(), parent_token->toAccessor());
        }
        return path;
    }

    // Containers
    void NavigatorWriter::writeStartDict()
    {
        Path path = generatePath();
        auto data = std::make_shared<datanavigator::DictNavigator>(path);
        auto new_token = std::make_shared<NavigatorWriterToken>(data::Descriptor::eDict, data);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndDict()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NavigatorWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartList()
    {
        Path path = generatePath();
        auto data = std::make_shared<datanavigator::ListNavigator>(path);
        auto new_token = std::make_shared<NavigatorWriterToken>(data::Descriptor::eList, data);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndList()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            NavigatorWriterTokenPtr prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    // Complex Types
    void NavigatorWriter::writeNDArray(const std::vector<int>& dims, const std::string& t, const unsigned char* data)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        datanavigator::NDArrayNavigatorPtr aron(new datanavigator::NDArrayNavigator(path));
        aron->setDimensions(dims);
        aron->setType(t);
        aron->setData(std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()), data);
        token->addElement(aron);
    }


    // Primitive Types
    void NavigatorWriter::writePrimitive(const int x)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto aron = std::make_shared<datanavigator::IntNavigator>(path);
        aron->setValue(x);
        token->addElement(aron);
    }

    void NavigatorWriter::writePrimitive(const long x)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto aron = std::make_shared<datanavigator::LongNavigator>(path);
        aron->setValue(x);
        token->addElement(aron);
    }

    void NavigatorWriter::writePrimitive(const float x)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto aron = std::make_shared<datanavigator::FloatNavigator>(path);
        aron->setValue(x);
        token->addElement(aron);
    }

    void NavigatorWriter::writePrimitive(const double x)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto aron = std::make_shared<datanavigator::DoubleNavigator>(path);
        aron->setValue(x);
        token->addElement(aron);
    }

    void NavigatorWriter::writePrimitive(const std::string& x)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto aron = std::make_shared<datanavigator::StringNavigator>(path);
        aron->setValue(x);
        token->addElement(aron);
    }

    void NavigatorWriter::writePrimitive(const bool x)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto aron = std::make_shared<datanavigator::BoolNavigator>(path);
        aron->setValue(x);
        token->addElement(aron);
    }

    void NavigatorWriter::writeKey(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }
}
