/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "Dict.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/navigator/data/NavigatorFactory.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Dict.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Object.h>

namespace armarx::aron::datanavigator
{

    // constructors
    DictNavigator::DictNavigator(const Path& path) :
        aron::Navigator<data::Descriptor, data::AronData>::Navigator(data::Descriptor::eDict, path),
        Navigator(data::Descriptor::eDict, path),
        aron(new data::AronDict())
    {
    }

    DictNavigator::DictNavigator(const data::AronDictPtr& o, const Path& path) :
        aron::Navigator<data::Descriptor, data::AronData>::Navigator(data::Descriptor::eDict, path),
        Navigator(data::Descriptor::eDict, path),
        aron(o)
    {
        CheckAronPtrForNull("DictNavigator", "DictNavigator", getPath(), aron);

        for (const auto& [key, dataPtr] : aron->elements)
        {
            childrenNavigators[key] = Navigator::FromAronData(dataPtr, Path(path, key));
        }
    }

    DictNavigator::DictNavigator(const data::AronDataDict& d, const Path& path) :
        DictNavigator(data::AronDictPtr(new data::AronDict(d)), path)
    {
    }

    DictNavigator::DictNavigator(const std::map<std::string, NavigatorPtr>& m, const Path& path) :
        DictNavigator(path)
    {
        for (const auto& [key, dataPtr] : m)
        {
            addElement(key, dataPtr);
        }
    }

    // operators
    bool DictNavigator::operator==(const DictNavigator& other) const
    {
        for (const auto& [key, nav] : childrenNavigators)
        {
            if (not(other.hasElement(key)))
            {
                return false;
            }
            if (not(nav->equalsDataNavigator(other.getElement(key))))
            {
                return false;
            }
        }
        return true;
    }

    bool DictNavigator::equalsDataNavigator(const NavigatorPtr& other) const
    {
        if (other == nullptr)
        {
            return false;
        }
        DictNavigatorPtr casted = DynamicCast(other);
        return equalsDeep(casted);
    }

    bool DictNavigator::equalsDeep(const DictNavigatorPtr& other) const
    {
        if (other == nullptr)
        {
            return false;
        }
        return *this == *other;
    }

    // static methods
    DictNavigatorPtr DictNavigator::DynamicCast(const NavigatorPtr& n)
    {
        return std::dynamic_pointer_cast<DictNavigator>(n);
    }

    DictNavigator DictNavigator::DynamicCast(Navigator& n)
    {
        return dynamic_cast<DictNavigator&>(n);
    }

    DictNavigatorPtr DictNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckDataNavigatorPtrForNull("DictNavigator", "DynamicCastAndCheck[Before]", n);
        DictNavigatorPtr casted = DictNavigator::DynamicCast(n);
        CheckDataNavigatorPtrForNull("DictNavigator", "DynamicCastAndCheck[After]", n->getPath(), casted);
        return casted;
    }

    DictNavigatorPtr DictNavigator::FromAronDictPtr(const data::AronDictPtr& aron)
    {
        return std::make_shared<DictNavigator>(aron);
    }

    data::AronDictPtr DictNavigator::ToAronDictPtr(const DictNavigatorPtr& navigator)
    {
        return navigator ? navigator->toAronDictPtr() : nullptr;
    }

    data::AronDictPtr DictNavigator::toAronDictPtr() const
    {
        return aron;
    }

    // public member functions
    std::vector<std::string> DictNavigator::getAllKeys() const
    {
        std::vector<std::string> ret;
        for (const auto& [key, _] : childrenNavigators)
        {
            ret.push_back(key);
        }
        return ret;
    }

    void DictNavigator::addElement(const std::string& key, const NavigatorPtr& data)
    {
        this->childrenNavigators[key] = data;
        this->aron->elements[key] = data->getResult();
    }

    bool DictNavigator::hasElement(const std::string& key) const
    {
        return childrenNavigators.count(key) > 0;
    }

    NavigatorPtr DictNavigator::getElement(const std::string& key) const
    {
        auto it = childrenNavigators.find(key);
        if (it == childrenNavigators.end())
        {
            std::string all_keys = "";
            for (const auto& child : this->getAllKeys())
            {
                all_keys += child + ", ";
            }
            throw error::AronException("DictNavigator", "getElement", "Could not find key '" + key + "'. But I found the following keys: [" + all_keys + "]", getPath());
        }
        return it->second;
    }

    std::map<std::string, NavigatorPtr> DictNavigator::getElements() const
    {
        return childrenNavigators;
    }

    void DictNavigator::clear()
    {
        childrenNavigators.clear();
        aron->elements.clear();
    }

    // virtual implementations
    data::AronDataPtr DictNavigator::getResult() const
    {
        return toAronDictPtr();
    }

    std::string DictNavigator::getName() const
    {
        return "AronDict";
    }

    typenavigator::NavigatorPtr DictNavigator::recalculateType() const
    {
        typenavigator::DictNavigatorPtr typenav = typenavigator::DictNavigatorPtr(new typenavigator::DictNavigator(getPath()));
        for (const auto& [key, nav] : childrenNavigators)
        {
            if (typenav->getAcceptedType() == nullptr)
            {
                typenav->setAcceptedType(nav->recalculateType());
                continue;
            }

            if (!nav->fullfillsType(typenav->getAcceptedType()))
            {
                throw error::AronException("DictNavigator", "recalculateType", "Cannot recalculate the aronType. Inconsistency found for key: " + key, getPath());
            }
        }
        return typenav->getAcceptedType();
    }

    bool DictNavigator::fullfillsType(const typenavigator::NavigatorPtr& type) const
    {
        if (!Resolver::Correspond(type->getDescriptor(), getDescriptor()))
        {
            return false;
        }

        type::Descriptor typeDesc = type->getDescriptor();
        switch (typeDesc)
        {
            case type::Descriptor::eObject:
            {
                typenavigator::ObjectNavigatorPtr objectTypeNav = typenavigator::ObjectNavigator::DynamicCast(type);
                for (const auto& [key, nav] : childrenNavigators)
                {
                    if (!nav->fullfillsType(objectTypeNav->getMemberType(key)))
                    {
                        return false;
                    }
                }
                return true;
            }
            case type::Descriptor::eDict:
            {
                typenavigator::DictNavigatorPtr dictTypeNav = typenavigator::DictNavigator::DynamicCast(type);
                for (const auto& [_, nav] : childrenNavigators)
                {
                    if (!nav->fullfillsType(dictTypeNav->getAcceptedType()))
                    {
                        return false;
                    }
                }
                return true;
            }
            default:
                return false;
        }
    }

    std::vector<NavigatorPtr> DictNavigator::getChildren() const
    {
        std::vector<NavigatorPtr> ret(childrenNavigators.size());
        for (const auto& [key, nav] : childrenNavigators)
        {
            ret.push_back(nav);
        }
        return ret;
    }

    size_t DictNavigator::childrenSize() const
    {
        return childrenNavigators.size();
    }

    NavigatorPtr DictNavigator::navigateAbsolute(const Path& path) const
    {
        if (!path.hasElement())
        {
            throw error::AronException("DictNavigator", "navigate", "Could not navigate without a valid path. The path was empty.");
        }
        std::string el = path.getFirstElement();
        if (!hasElement(el))
        {
            throw error::StringNotValidException("DictNavigator", "navigate", "Could not find an element of a path.", el);
        }

        if (path.size() == 1)
        {
            return childrenNavigators.at(el);
        }
        else
        {
            Path next = path.withDetachedFirstElement();
            return childrenNavigators.at(el)->navigateAbsolute(next);
        }
    }

    NavigatorPtr DictNavigator::navigateRelative(const Path& path) const
    {
        Path absoluteFromHere = path.getWithoutPrefix(getPath());
        return navigateAbsolute(absoluteFromHere);
    }
}
